"""
Wrapper for iblatlas
======================

Provide basic plotting usage

See detail in ``iblatlas.plots.plot_scalar_on_slice``

.. seealso:: `<https://int-brain-lab.github.io/iblenv/notebooks_external/atlas_plotting_scalar_on_slice.html>`_


.. toctree::
    :maxdepth: 1
    :caption: examples

    ../notebooks/example_ibl_plot

"""

from .plot import *
